#define FREQUENCY_868

#include <Arduino.h>
#include <TM1637.h>
#include <Ticker.h>
#include <DNSServer.h>
#include <ESP8266WiFi.h>
#include <ESPAsyncTCP.h>
#include <ESP8266mDNS.h>
#include <ArduinoOTA.h>
#include <ESPAsyncWebServer.h>
#include <LittleFS.h>
#include <PCF8574.h>
#include <Wire.h>
#include <Adafruit_INA219.h>
#include <ArduinoJson.h>
#include <LoRa_E220.h>

// local includes
#include "lora_messages.h"
#include "defaults.h"
#include "webui.h"
#include "config.h"
#include "globals.h"
#include "dtc.h"
#include "common.h"

#ifdef WIFI_CLIENT
#include <WiFiMulti.h>

const char *ssid = QUOTE(WIFI_SSID);
const char *password = QUOTE(WIFI_PASSWORD);
const uint32_t connectTimeoutMs = 5000;

WiFiMulti wifiMulti;
#endif

PCF8574 i2c_io(I2C_IO_ADDRESS);
Adafruit_INA219 ina219;
LoRa_E220 e220ttl(GPIO_LORA_TX, GPIO_LORA_RX, GPIO_LORA_AUX, 3, 4); // Arduino RX <-- e220 TX, Arduino TX --> e220 RX AUX M0 M1

TM1637 disp_FAC_1(GPIO_7SEG_CLK, GPIO_7SEG_EN_FAC1);
TM1637 disp_FAC_2(GPIO_7SEG_CLK, GPIO_7SEG_EN_FAC2);
TM1637 disp_FAC_3(GPIO_7SEG_CLK, GPIO_7SEG_EN_FAC3);

#ifdef CAPTIVE
DNSServer dnsServer;
#endif
AsyncWebServer server(80);

void printParameters(struct Configuration configuration);
void printModuleInformation(struct ModuleInformation moduleInformation);
void displayInfo();
void SevenSeg_Output();
void toggleWiFiAP(boolean shutdown = false);
void SystemShutdown();
void SetBatteryType(batteryType_t type);
void ProcessKeyCombos(bool *btnState);
void OverrideDisplay(const uint8_t *message, uint32_t time);
uint32_t getESPChipID();

void tmrCallback_StatusSender();
Ticker tmrStatusSender(tmrCallback_StatusSender, 30000, 0, MILLIS);
void tmrCallback_PowerMonitor();
Ticker tmrPowerMonitor(tmrCallback_PowerMonitor, 10000, 0, MILLIS);
void tmrCallback_FactionTicker();
Ticker tmrFactionTicker(tmrCallback_FactionTicker, 1000, 0, MILLIS);
void tmrCallback_InputGetter();
Ticker tmrInputGetter(tmrCallback_InputGetter, 250, 0, MILLIS);
void tmrCallback_EEPROMCyclicPDS();
Ticker tmrEEPROMCyclicPDS(tmrCallback_EEPROMCyclicPDS, 60000, 0, MILLIS);

#ifdef WIFI_CLIENT
void tmrCallback_WiFiMaintainConnection();
Ticker tmrWiFiMaintainConnection(tmrCallback_WiFiMaintainConnection, 1000, 0, MILLIS);
#endif

uint32_t DisplayOverrideFlag = 0;
char DisplayOverrideValue[5] = {0};

Globals_t globals;

void setMPins_Helper(int pin, int status)
{
	i2c_io.write(pin, status);
}

void setup()
{
	system_update_cpu_freq(SYS_CPU_80MHZ);

	Serial.begin(115200);
	Serial.print("\n\n\n");

#ifdef SERIAL_DEBUG
	Serial.setDebugOutput(true);
#endif

	WiFi.persistent(false);
	WiFi.disconnect();

	strcpy(globals.DeviceName, DEVICE_NAME);
	snprintf(globals.DeviceName_ID, 42, "%s_%08X", globals.DeviceName, getESPChipID());

	if (i2c_io.begin())
	{
		Serial.printf("Initialized PCF8574-GPIO at Address 0x%02X\n", I2C_IO_ADDRESS);
	}
	else
	{
		Serial.print("PCF8574-GPIO not Initialized\n");
	}

	if (ina219.begin())
	{
		Serial.printf("Initialized INA219-Powermonitor at Address 0x%02X\n", I2C_POWER_ADDRESS);
		tmrPowerMonitor.start();
	}
	else
	{
		Serial.print("INA219 not Initialized\n");
	}

	e220ttl.setMPins = &setMPins_Helper;
	e220ttl.begin();

	ResponseStructContainer c;
	c = e220ttl.getConfiguration();
	// It's important get configuration pointer before all other operation
	Configuration configuration = *(Configuration *)c.data;
	Serial.println(c.status.getResponseDescription());
	Serial.println(c.status.code);

	ResponseStructContainer cMi;
	cMi = e220ttl.getModuleInformation();
	// It's important get information pointer before all other operation
	// ModuleInformation mi = *(ModuleInformation *)cMi.data;

	Serial.println(cMi.status.getResponseDescription());
	Serial.println(cMi.status.code);

	//	----------------------- DEFAULT TRANSPARENT WITH RSSI -----------------------
	configuration.ADDL = 0x02;
	configuration.ADDH = 0x00;

	configuration.CHAN = 23;

	configuration.SPED.uartBaudRate = UART_BPS_9600;
	configuration.SPED.airDataRate = AIR_DATA_RATE_010_24;
	configuration.SPED.uartParity = MODE_00_8N1;

	configuration.OPTION.subPacketSetting = SPS_200_00;
	configuration.OPTION.RSSIAmbientNoise = RSSI_AMBIENT_NOISE_ENABLED;
	configuration.OPTION.transmissionPower = POWER_22;

	configuration.TRANSMISSION_MODE.enableRSSI = RSSI_ENABLED;
	configuration.TRANSMISSION_MODE.fixedTransmission = FT_FIXED_TRANSMISSION;
	configuration.TRANSMISSION_MODE.enableLBT = LBT_ENABLED;
	configuration.TRANSMISSION_MODE.WORPeriod = WOR_2000_011;

	// Set configuration changed and set to not hold the configuration
	ResponseStatus rs = e220ttl.setConfiguration(configuration, WRITE_CFG_PWR_DWN_LOSE);
	Serial.println(rs.getResponseDescription());
	Serial.println(rs.code);
	c.close();

	printParameters(configuration);
	tmrStatusSender.start();

	LittleFS.begin();

#ifdef WIFI_CLIENT
	WiFi.mode(WIFI_STA);
	WiFi.setHostname(globals.DeviceName_ID);
	wifiMulti.addAP(QUOTE(WIFI_SSID), QUOTE(WIFI_PASSWORD));
	tmrWiFiMaintainConnection.start();
#else
	WiFi.mode(WIFI_AP);
	WiFi.begin(QUOTE(DEVICE_NAME), QUOTE(WIFI_AP_PASSWORD));
#if defined(ESP32)
	WiFi.sleep(true);
#endif
	WiFi.mode(WIFI_OFF);
#endif

	InitEEPROM();

	ArduinoOTA.setPort(8266);
	ArduinoOTA.setHostname(globals.DeviceName_ID);
	ArduinoOTA.setPassword(QUOTE(ADMIN_PASSWORD));

	ArduinoOTA.onStart([]()
					   {
                       disp_FAC_1.setBrightness(5);
                       disp_FAC_2.setBrightness(5);
                       disp_FAC_3.setBrightness(5);
                       disp_FAC_1.display("OTA ");
                       disp_FAC_3.clearScreen();
                       
                       if (ArduinoOTA.getCommand() == U_FLASH)
                       {
                         disp_FAC_2.display("FLSH");
                       }
                       else
                       {
                         disp_FAC_2.display("FILE");
                         LittleFS.end();
                       } });

	ArduinoOTA.onEnd([]()
					 { 
                       disp_FAC_1.display("DONE");
                       disp_FAC_2.clearScreen();
                       disp_FAC_3.clearScreen(); });

	ArduinoOTA.onProgress([](unsigned int progress, unsigned int total)
						  { disp_FAC_3.display((progress / (total / 100))); });

	ArduinoOTA.onError([](ota_error_t error)
					   {
                       Serial.printf("Error[%u]: ", error);
                       if (error == OTA_AUTH_ERROR)
                         Serial.println("Auth Failed");
                       else if (error == OTA_BEGIN_ERROR)
                         Serial.println("Begin Failed");
                       else if (error == OTA_CONNECT_ERROR)
                         Serial.println("Connect Failed");
                       else if (error == OTA_RECEIVE_ERROR)
                         Serial.println("Receive Failed");
                       else if (error == OTA_END_ERROR)
                         Serial.println("End Failed"); });
	ArduinoOTA.begin();

#ifdef CAPTIVE
	dnsServer.start(53, "*", WiFi.softAPIP());
#endif

	initWebUI();

	disp_FAC_1.init();
	disp_FAC_1.setBrightness(5);
	disp_FAC_2.init();
	disp_FAC_2.setBrightness(5);
	disp_FAC_3.init();
	disp_FAC_3.setBrightness(5);

	tmrEEPROMCyclicPDS.start();
	tmrFactionTicker.start();
	tmrInputGetter.start();
	Serial.println("Setup Done");
}

void loop()
{

	if (e220ttl.available() > 1)
	{
		ResponseContainer rc = e220ttl.receiveMessageRSSI();
		// Is something goes wrong print error
		if (rc.status.code != 1)
		{
			Serial.println(rc.status.getResponseDescription());
		}
		else
		{
			// Print the data received
			Serial.println(rc.status.getResponseDescription());
			Serial.println(rc.data);
			Serial.print("RSSI: ");
			Serial.println(rc.rssi, DEC);
		}
	}

	tmrEEPROMCyclicPDS.update();
	tmrFactionTicker.update();
	tmrInputGetter.update();
	tmrStatusSender.update();
	tmrPowerMonitor.update();

	ArduinoOTA.handle();
	SevenSeg_Output();
	EEPROM_Process();

#ifdef CAPTIVE
	dnsServer.processNextRequest();
#endif
#ifdef WIFI_CLIENT
	tmrWiFiMaintainConnection.update();
#endif
	if (globals.systemStatus == sysStat_Shutdown)
		SystemShutdown();

	yield();
}

String macToString(const unsigned char *mac)
{
	char buf[20];
	snprintf(buf, sizeof(buf), "%02x:%02x:%02x:%02x:%02x:%02x",
			 mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
	return String(buf);
}

void SevenSeg_Output()
{
	char sevenSegBuff[5] = "";

	if (DisplayOverrideFlag > millis())
	{
		disp_FAC_1.setBrightness(5);
		disp_FAC_2.setBrightness(5);
		disp_FAC_3.setBrightness(5);
		disp_FAC_1.display(String(DisplayOverrideValue));
		disp_FAC_2.clearScreen();
		disp_FAC_3.clearScreen();
	}
	else
	{
		if (globals.battery_level < BAT_LOW_PERCENT && millis() % 10000 > 7000)
		{
			if (millis() % 3000 < 1500)
				snprintf(sevenSegBuff, sizeof(sevenSegBuff), "%4d", globals.battery_level);
			else
				snprintf(sevenSegBuff, sizeof(sevenSegBuff), "%4s", String(globals.loadvoltage, 1).c_str());

			disp_FAC_1.setBrightness(1);
			disp_FAC_1.display(" BAT");

			disp_FAC_2.setBrightness(1);
			disp_FAC_2.display("LOW ");

			disp_FAC_3.setBrightness(1);
			disp_FAC_3.display(String(sevenSegBuff));
		}
		else
		{
			disp_FAC_1.setBrightness(PersistenceData.activeFaction == FACTION_1 ? 5 : 1);
			disp_FAC_1.refresh();
			snprintf(sevenSegBuff, sizeof(sevenSegBuff), "%4d", PersistenceData.faction_1_timer / 60);
			disp_FAC_1.display(String(sevenSegBuff), false, false);

			disp_FAC_2.setBrightness(PersistenceData.activeFaction == FACTION_2 ? 5 : 1);
			disp_FAC_2.refresh();
			snprintf(sevenSegBuff, sizeof(sevenSegBuff), "%4d", PersistenceData.faction_2_timer / 60);
			disp_FAC_2.display(String(sevenSegBuff), false, false);

			disp_FAC_3.setBrightness(PersistenceData.activeFaction == FACTION_3 ? 5 : 1);
			disp_FAC_3.refresh();
			snprintf(sevenSegBuff, sizeof(sevenSegBuff), "%4d", PersistenceData.faction_3_timer / 60);
			disp_FAC_3.display(String(sevenSegBuff), false, false);
		}
	}
}

void tmrCallback_FactionTicker()
{
	switch (PersistenceData.activeFaction)
	{
	case FACTION_1:
		PersistenceData.faction_1_timer++;
		break;

	case FACTION_2:
		PersistenceData.faction_2_timer++;
		break;

	case FACTION_3:
		PersistenceData.faction_3_timer++;
		break;

	default:
		break;
	}
}

void tmrCallback_InputGetter()
{
	static bool btnState[3];
	uint8_t keysPressed = 0;

	btnState[0] = i2c_io.readButton(I2C_IO_BTN_FAC1);
	btnState[1] = i2c_io.readButton(I2C_IO_BTN_FAC2);
	btnState[2] = i2c_io.readButton(I2C_IO_BTN_FAC3);

	ProcessKeyCombos(btnState);

	keysPressed = keysPressed + (btnState[0] == FAC_1_TRG_PRESSED ? 1 : 0);
	keysPressed = keysPressed + (btnState[1] == FAC_2_TRG_PRESSED ? 1 : 0);
	keysPressed = keysPressed + (btnState[2] == FAC_3_TRG_PRESSED ? 1 : 0);

	if (keysPressed > 1)
	{
		Serial.println("ERROR: More than one Flag active - setting no Faction active");
		PersistenceData.activeFaction = NONE;
		return;
	}

	if (btnState[0] == FAC_1_TRG_PRESSED)
	{
		if (PersistenceData.activeFaction != FACTION_1)
		{
			Serial.println("Faction 1 captured !");
		}
		PersistenceData.activeFaction = FACTION_1;
	}

	if (btnState[1] == FAC_2_TRG_PRESSED)
	{
		if (PersistenceData.activeFaction != FACTION_2)
		{
			Serial.println("Faction 2 captured !");
		}
		PersistenceData.activeFaction = FACTION_2;
	}

	if (btnState[2] == FAC_3_TRG_PRESSED)
	{
		if (PersistenceData.activeFaction != FACTION_3)
		{
			Serial.println("Faction 3 captured !");
		}
		PersistenceData.activeFaction = FACTION_3;
	}
}

void printParameters(struct Configuration configuration)
{
	Serial.println("----------------------------------------");

	Serial.print(F("HEAD : "));
	Serial.print(configuration.COMMAND, HEX);
	Serial.print(" ");
	Serial.print(configuration.STARTING_ADDRESS, HEX);
	Serial.print(" ");
	Serial.println(configuration.LENGHT, HEX);
	Serial.println(F(" "));
	Serial.print(F("AddH : "));
	Serial.println(configuration.ADDH, HEX);
	Serial.print(F("AddL : "));
	Serial.println(configuration.ADDL, HEX);
	Serial.println(F(" "));
	Serial.print(F("Chan : "));
	Serial.print(configuration.CHAN, DEC);
	Serial.print(" -> ");
	Serial.println(configuration.getChannelDescription());
	Serial.println(F(" "));
	Serial.print(F("SpeedParityBit     : "));
	Serial.print(configuration.SPED.uartParity, BIN);
	Serial.print(" -> ");
	Serial.println(configuration.SPED.getUARTParityDescription());
	Serial.print(F("SpeedUARTDatte     : "));
	Serial.print(configuration.SPED.uartBaudRate, BIN);
	Serial.print(" -> ");
	Serial.println(configuration.SPED.getUARTBaudRateDescription());
	Serial.print(F("SpeedAirDataRate   : "));
	Serial.print(configuration.SPED.airDataRate, BIN);
	Serial.print(" -> ");
	Serial.println(configuration.SPED.getAirDataRateDescription());
	Serial.println(F(" "));
	Serial.print(F("OptionSubPacketSett: "));
	Serial.print(configuration.OPTION.subPacketSetting, BIN);
	Serial.print(" -> ");
	Serial.println(configuration.OPTION.getSubPacketSetting());
	Serial.print(F("OptionTranPower    : "));
	Serial.print(configuration.OPTION.transmissionPower, BIN);
	Serial.print(" -> ");
	Serial.println(configuration.OPTION.getTransmissionPowerDescription());
	Serial.print(F("OptionRSSIAmbientNo: "));
	Serial.print(configuration.OPTION.RSSIAmbientNoise, BIN);
	Serial.print(" -> ");
	Serial.println(configuration.OPTION.getRSSIAmbientNoiseEnable());
	Serial.println(F(" "));
	Serial.print(F("TransModeWORPeriod : "));
	Serial.print(configuration.TRANSMISSION_MODE.WORPeriod, BIN);
	Serial.print(" -> ");
	Serial.println(configuration.TRANSMISSION_MODE.getWORPeriodByParamsDescription());
	Serial.print(F("TransModeEnableLBT : "));
	Serial.print(configuration.TRANSMISSION_MODE.enableLBT, BIN);
	Serial.print(" -> ");
	Serial.println(configuration.TRANSMISSION_MODE.getLBTEnableByteDescription());
	Serial.print(F("TransModeEnableRSSI: "));
	Serial.print(configuration.TRANSMISSION_MODE.enableRSSI, BIN);
	Serial.print(" -> ");
	Serial.println(configuration.TRANSMISSION_MODE.getRSSIEnableByteDescription());
	Serial.print(F("TransModeFixedTrans: "));
	Serial.print(configuration.TRANSMISSION_MODE.fixedTransmission, BIN);
	Serial.print(" -> ");
	Serial.println(configuration.TRANSMISSION_MODE.getFixedTransmissionDescription());

	Serial.println("----------------------------------------");
}

void tmrCallback_StatusSender()
{
	struct
	{
		MessageType_t type = "STATUS";
		MessageStatus_t status;
	} __attribute__((packed)) sendStatus;

	sendStatus.status.nodeid = 0x0002;
	sendStatus.status.millis = millis();
	sendStatus.status.faction_active = 1;
	sendStatus.status.faction_1_timer = 0xBBBBBBBB;
	sendStatus.status.faction_2_timer = 0xCCCCCCCC;
	sendStatus.status.faction_3_timer = 0xDDDDDDDD;

	ResponseStatus rs = e220ttl.sendFixedMessage(0xFF, 0xFF, 23, (byte *)&sendStatus, sizeof(sendStatus));
	Serial.println(rs.getResponseDescription());
}

void tmrCallback_PowerMonitor()
{
	// loadvoltage and percentage is global, because of battery Monitoring

	const int bat_min_2s = 680;
	const int bat_max_2s = 840;
	const int bat_min_3s = 1020;
	const int bat_max_3s = 1260;

	float shuntvoltage = 0;
	// float current_mA = 0;
	float busvoltage = 0;
	// float power_mW = 0;
	int battery_level = 0;

	shuntvoltage = ina219.getShuntVoltage_mV();
	busvoltage = ina219.getBusVoltage_V();
	// current_mA = ina219.getCurrent_mA();
	// power_mW = ina219.getPower_mW();
	globals.loadvoltage = busvoltage + (shuntvoltage / 1000);
	switch (ConfigData.batteryType)
	{
	case BATTERY_LIPO_2S:
		battery_level = map(globals.loadvoltage * 100, bat_min_2s, bat_max_2s, 0, 100);
		globals.battery_level = battery_level < 0 ? 0 : battery_level;
		break;
	case BATTERY_LIPO_3S:
		battery_level = map(globals.loadvoltage * 100, bat_min_3s, bat_max_3s, 0, 100);
		globals.battery_level = battery_level < 0 ? 0 : battery_level;
		break;
	default:
		globals.battery_level = -1;
		break;
	}

	// Serial.printf("Battery Level: %d %%\n", globals.battery_level);
	// Serial.printf("Bus Voltage: %f V\n", busvoltage);
	// Serial.printf("Shunt Voltage: %f mV\n", shuntvoltage);
	// Serial.printf("Load Voltage: %f V\n", globals.loadvoltage);
	// Serial.printf("Current: %f mA\n", current_mA);
	// Serial.printf("Power: %f mW\n", power_mW);
}

void tmrCallback_EEPROMCyclicPDS()
{
	StorePersistence_EEPROM();
}

#ifdef WIFI_CLIENT
void tmrCallback_WiFiMaintainConnection()
{
	static uint32_t WiFiFailCount = 0;
	const uint32_t WiFiFailMax = 20;

	if (wifiMulti.run(connectTimeoutMs) == WL_CONNECTED)
	{
		return;
	}
	else
	{
		if (WiFiFailCount < WiFiFailMax)
			WiFiFailCount++;
		else
			toggleWiFiAP(false);
	}
}
#endif

void toggleWiFiAP(boolean shutdown)
{
	if (WiFi.getMode() != WIFI_OFF && shutdown == true)
	{
		WiFi.mode(WIFI_OFF);
#ifdef WIFI_CLIENT
		tmrWiFiMaintainConnection.stop();
#endif
	}
	else if (shutdown == false)
	{
		WiFi.mode(WIFI_AP);
		WiFi.softAPConfig(IPAddress(WIFI_AP_IP_GW), IPAddress(WIFI_AP_IP_GW), IPAddress(255, 255, 255, 0));
		WiFi.softAP(globals.DeviceName_ID, QUOTE(WIFI_AP_PASSWORD));
#ifdef WIFI_CLIENT
		tmrWiFiMaintainConnection.stop();
#endif
	}
}

void SystemShutdown()
{
	StoreConfig_EEPROM();
	ESP.restart();
}

void SetBatteryType(batteryType_t type)
{
	if (ConfigData.batteryType != type)
	{
		ConfigData.batteryType = type;
		globals.requestEEAction = EE_CFG_SAVE;
		Serial.printf("Set Batterytype to %s\n", type == BATTERY_LIPO_2S ? "2s Lipo" : "3s LiPo");
	}
}

void OverrideDisplay(const char *message, uint32_t time)
{
	DisplayOverrideFlag = millis() + time;
	strcpy(DisplayOverrideValue, message);
}

uint32_t getESPChipID()
{
#if defined(ESP8266)
	return ESP.getChipId();
#elif defined(ESP32)
	uint32_t chipId;
	for (int i = 0; i < 17; i = i + 8)
	{
		chipId |= ((ESP.getEfuseMac() >> (40 - i)) & 0xff) << i;
	}
	return chipId;
#endif
}

void ProcessKeyCombos(bool *btnState)
{
	typedef enum
	{
		KEY_PRESSED,
		KEY_RELEASED
	} keyStatus_t;

	static keyStatus_t keyStatus_Fac1 = KEY_RELEASED;
	static uint8_t keyCount_Fac1 = 0;
	static keyStatus_t keyStatus_Fac2 = KEY_RELEASED;
	static uint8_t keyCount_Fac2 = 0;
	static keyStatus_t keyStatus_Fac3 = KEY_RELEASED;

	if (btnState[2] == FAC_3_TRG_PRESSED)
	{
		keyStatus_Fac3 = KEY_PRESSED;

		// Process FactionKey 1 ComboCounter
		if (btnState[0] == FAC_1_TRG_PRESSED && keyStatus_Fac1 == KEY_RELEASED)
		{
			keyStatus_Fac1 = KEY_PRESSED;
			keyCount_Fac1++;
		}

		if (btnState[0] != FAC_1_TRG_PRESSED)
			keyStatus_Fac1 = KEY_RELEASED;

		// Process FactionKey 2 ComboCounter
		if (btnState[1] == FAC_2_TRG_PRESSED && keyStatus_Fac2 == KEY_RELEASED)
		{
			keyStatus_Fac2 = KEY_PRESSED;
			keyCount_Fac2++;
		}

		if (btnState[1] != FAC_2_TRG_PRESSED)
			keyStatus_Fac2 = KEY_RELEASED;
	}

	if (btnState[2] != FAC_3_TRG_PRESSED && keyStatus_Fac3 == KEY_PRESSED)
	{
		if (keyCount_Fac1 > 0 || keyCount_Fac2 > 0)
			Serial.printf("KeyCombo 1: %d | 2: %d\n", keyCount_Fac1, keyCount_Fac2);

		if (keyCount_Fac1 == 2 && keyCount_Fac2 == 2)
		{
			Serial.println("KeyCombo: WiFi AP ON");
			OverrideDisplay("NET ", 5000);
			toggleWiFiAP(false);
		}

		keyCount_Fac1 = 0;
		keyCount_Fac2 = 0;
		keyStatus_Fac1 = KEY_RELEASED;
		keyStatus_Fac2 = KEY_RELEASED;
		keyStatus_Fac3 = KEY_RELEASED;
	}
}